#pragma once

// C++ system level
#include <set>
#include <cstring>     // memset, strncmp
// #include <cstdlib>     // TODO: is this for memcpy?
#include <cinttypes>
// #include <functional>

// ESP32 specific
// #include "esp_err.h"
#include "nvs_flash.h"
#include "esp_log.h"
// #include "driver/gpio.h"
// #include "driver/uart.h"

// project specific
#include "SpiderLib/NumberTypes.hpp"
// #include "SpiderLib/ManagedMutex.hpp"

// qthing stuff
// #include <qthing>


namespace SiliconTorch {

  namespace NVSExplorer {


    extern const char* TAG;


    class NVSExplorer {

      public:

        std::set<std::string> listNamespaces();
        std::set<std::string> listKeys(const std::string& nameSpace);

        void rmKey(const std::string& nameSpace, const std::string& key);


        bool setU8(const  std::string& nameSpace, const std::string& key, u8  value);
        bool setI8(const  std::string& nameSpace, const std::string& key, i8  value);

        bool setU16(const std::string& nameSpace, const std::string& key, u16 value);
        bool setI16(const std::string& nameSpace, const std::string& key, i16 value);

        bool setU32(const std::string& nameSpace, const std::string& key, u32 value);
        bool setI32(const std::string& nameSpace, const std::string& key, i32 value);

        bool setU64(const std::string& nameSpace, const std::string& key, u64 value);
        bool setI64(const std::string& nameSpace, const std::string& key, i64 value);


        u8  getU8(const  std::string& nameSpace, const std::string& key, u8  defaultValue = 0);
        i8  getI8(const  std::string& nameSpace, const std::string& key, i8  defaultValue = 0);

        u16 getU16(const std::string& nameSpace, const std::string& key, u16 defaultValue = 0);
        i16 getI16(const std::string& nameSpace, const std::string& key, i16 defaultValue = 0);

        u32 getU32(const std::string& nameSpace, const std::string& key, u32 defaultValue = 0);
        i32 getI32(const std::string& nameSpace, const std::string& key, i32 defaultValue = 0);

        u64 getU64(const std::string& nameSpace, const std::string& key, u64 defaultValue = 0);
        i64 getI64(const std::string& nameSpace, const std::string& key, i64 defaultValue = 0);


        float  getFloat(const std::string& nameSpace, const std::string& key);
        double getDouble(const std::string& nameSpace, const std::string& key);

        bool setFloat(const std::string& nameSpace, const std::string& key, float value);
        bool setDouble(const std::string& nameSpace, const std::string& key, double value);

        // Gets the key regardless of its type in flash
        i64  getSignedInt(const std::string& nameSpace, const std::string& key, i64 defaultValue = 0);
        u64 getUnsignedInt(const std::string& nameSpace, const std::string& key, u64 defaultValue = 0);

        // TODO: ⮦ Both unimplemented atm ⮧ 
        bool setSignedInt(const std::string& nameSpace, const std::string& key, i64 value, nvs_type_t entryType = NVS_TYPE_ANY);
        bool setUnsignedInt(const std::string& nameSpace, const std::string& key, u64 value, nvs_type_t entryType = NVS_TYPE_ANY);


        size_t getStringLength(const std::string& nameSpace, const std::string& key, size_t defaultValue = 0);
        std::string getString(const std::string& nameSpace, const std::string& key, const std::string& defaultValue = "");
        // size_t getLongString(const std::string& nameSpace, const std::string& key, char* const buffer);

        bool setString(const std::string& nameSpace, const std::string& key, const std::string& value);


        std::string type2str(nvs_type_t typ);
        nvs_type_t getType(const std::string& nameSpace, const std::string& key);
        std::string getTypeStr(const std::string& nameSpace, const std::string& key);


        std::string deviceTopic(const std::string& suffix);

        std::string requestTopic(const std::string& cmd);
        std::string requestTopic(const std::string& cmd, const std::string& nameSpace);
        std::string requestTopic(const std::string& cmd, const std::string& nameSpace, const std::string& key);

        std::string responseTopic(const std::string& cmd);
        std::string responseTopic(const std::string& cmd, const std::string& nameSpace);
        std::string responseTopic(const std::string& cmd, const std::string& nameSpace, const std::string& key);

        // ⬇⬇ Singleton stuff ⬇⬇

        static NVSExplorer& instance() {
          static NVSExplorer _instance;
          return _instance;
        }

        NVSExplorer(NVSExplorer const&)    = delete;
        void operator=(NVSExplorer const&) = delete;

        // ⬆⬆ Singleton stuff ⬆⬆

      private:

        void registerLambdas0();
        void registerLambdas1();

        NVSExplorer();

    };


  }
}
